class UnknownTarget(Exception):
    """Target state can not be found"""


class UnknownAction(Exception):
    """Action could not be found"""


class UnknownGuard(Exception):
    """Guard could not be found"""
